from decimal import Decimal


def remove_exponent(d: Decimal):
    """
    去除十进制小数（Decimal）的尾导零。

    非原创代码，出自：
    https://docs.python.org/zh-cn/3/library/decimal.html#decimal-faq
    """
    return d.quantize(Decimal(1)) if d == d.to_integral() else d.normalize()



import sys
from datetime import timedelta, datetime
from functools import wraps
from io import TextIOWrapper


def delta2hms(delta: timedelta) -> tuple:
    """
    将时间增量转换为时分秒格式。

    :param delta: 时间增量。
    :return: 包含三个整数的元组，分别对应小时数、分钟数、秒钟数。
    """
    h = delta.seconds // 3600
    m = delta.seconds % 3600 // 60
    s = delta.seconds % 60 + delta.microseconds / 1000000
    return h, m, s


def delta2ms(delta: timedelta) -> tuple:
    """
    将时间增量转换为分秒格式。

    :param delta: 时间增量。
    :return: 包含两个整数的元组，分别对应分钟数和秒钟数。
    """
    m = delta.seconds // 60
    s = delta.seconds % 60 + delta.microseconds / 1000000
    return m, s


class BearTimer(object):
    """
    熊牌计时器。

    这个类可以让你对代码进行计时功能并输出提示。

    最简单的用法是使用 with 语句包裹需要计时的代码：
    >>> with BearTimer():
    >>>     for i in range(1000000):
    >>>         print(i)

    如果不方便使用 with 包裹，可以直接实例化一个类对象：
    >>> bear = BearTimer()
    >>> bear.start()
    >>> bear.stop()

    计时开始后，如果需要打印提示，请使用 .step() 方法，例如：
    >>> with BearTimer() as bear:
    >>>     summary = 0
    >>>     for i in range(1000000):
    >>>         bear.step(f'loop to {i} now.')
    >>>         summary += i
    又或者是
    >>> bear = BearTimer()
    >>> bear.start()
    >>> bear.step('operate somethings now.')
    >>> bear.stop()
    """
    file: TextIOWrapper = sys.stdout

    @staticmethod
    def get_context_name():
        try:
            return sys._getframe().f_back.f_code.co_name
        except AttributeError:
            raise RuntimeError(
                'Current system NOT support this class.\n'
                '当前操作系统不支持使用该工具类。'
            ) from None

    def __init__(self, title: str = None, file: TextIOWrapper = None):
        """
        :param title: 计时器的标题，用以标明输出信息归属于哪个计时器。默认从上下文中获取。
        :param file: 消息打印到哪里去。默认是系统标准输出。
        :raise RuntimeError: 当前操作系统不支持使用该工具类。仅当没有提供title且无法从上下文中获取时抛出。
        """
        self._start = None  # 计时开始时间
        self._point = None  # 中途标记时间（用于计算距离上次标记过去了多久）
        self._title = title
        self.file = self.file if file is None else file

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self._title = func.__name__
            self.start()
            returns = func(*args, **kwargs)
            self.stop()
            return returns

        return wrapper

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.stop()

    def output(
            self,
            msg: str = '',
            fmt: str = '{head:%H:%M:%S.%f}, {minutes:d}:{seconds:f}, {title} | {msg}',
            *args,
            end: str = '\n',
            **kwargs,
    ) -> datetime:
        """
        打印一行消息，标明此刻的时间、经历的时长、计时器标题，以及附加的消息。

        你不应该使用这个方法，它只应在自定义这个类时被重写。

        :param msg: 要附加的消息。默认为空文本。
        :param fmt: 字符串的格式。具体参数见默认值。
        :param end: 要以什么结尾。因为一般是单行打印的，故有此参数。默认是 ”\n“ 。
        :param args: 其它需要打印的位置参数。
        :param kwargs: 其它需要打印的关键字参数。
        :return: 此时的时刻。
        """
        now = datetime.now()
        m, s = delta2ms(now - self._start) if self._start else (0, 0)
        title = self._title if self._title else self.get_context_name()
        print(
            fmt.format(*args, head=now, minutes=m, seconds=s,
                       title=title, msg=msg, **kwargs),
            file=self.file,
            end=end,
        )
        return now

    def start(self, msg='Starting...'):
        """
        开始计时。

        如果计时器还没结束，会重新开始计时，中途的标记也会被清除。

        :param msg: 要附加的消息。
        :return: 计时器对象自身。
        """
        self._point = None
        self._start = self.output(msg + ('＜计时已重新开始＞' if self._start else ''))
        return self

    def step(self, msg=None):
        """
        中途标记。

        用此方法可以计算出距离上一次标记/开始计时过去了多久。

        如果计时器尚未开始计时，那么过去的时间始终是 0 秒。

        :param msg: 要附加的消息。留空则显示过了多久。
        :return: 一个元组，包含上一次标记/开始计时的时刻，和此时的时刻。
        """
        if self._start is not None:
            if self._point is None:
                self._point = self._start
            m, s = delta2ms(datetime.now() - self._point)
        else:
            m, s = 0, 0
        prev = self._point
        curr = self.output(msg if msg else f'(+{m:d}:{s:f})')
        self._point = curr
        return prev, curr

    def stop(self, msg='Stopped.') -> datetime:
        """
        结束计时。

        :param msg: 要附加的消息。
        :return: 此时的时刻。
        """
        curr = self.output(msg + ('' if self._start else '＜计时还未开始＞'))
        self._start = None
        return curr



from .decorators import start
from .generators import (
    randbytes,
    randb62,
    randb64,
    SnowflakeWorker,
    SnowflakeMultiWorker,
    SnowflakeSingleWorker,
)
from .math import remove_exponent
from .time import (
    delta2hms,
    delta2ms,
    BearTimer,
)
from .typing import (
    JSONObject,
    JsonObject,
    casting,
    ReprMixin,
)

__author__ = 'aixcyi'
__version__ = (0, 1, 1)
version = '0.1.1'



import sys
from functools import wraps


def start(*version: int, note=None):
    """
    检查 Python 版本是否高于或等于指定值，
    如果低于指定的版本就会抛出 RuntimeError。
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if sys.version_info < version:
                raise RuntimeError(
                    f'Require Python version {version:s} to run. '
                    f'运行此函数/类需要Python版本在 {version:s} 及以上。'
                    + ('' if note is None else note)
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator



from typing import Callable


class JSONObject(object):
    def __init__(self, dictionary: dict = None, **kvs):
        """
        将嵌套字典转化为嵌套对象（递归转化）。

        :param dictionary: 包含数据的字典。请确保所有键都符合标识符命名要求，且均为字符串。
        :param kvs: 任意键值对。
        :return: JSONObject。
        :raise TypeError: 属性名称必须是字符串。
        """
        for k, v in {**(dictionary if dictionary else {}), **kvs}.items():
            k = str(k)
            if not k.isidentifier():
                continue
            if v.__class__ is list:
                self.__setattr__(k, self._translate_list(v[:]))
            elif v.__class__ is dict:
                self.__setattr__(k, self.__class__(v))
            else:
                self.__setattr__(k, v)

    def __or__(self, dictionary: dict):
        for k, v in dictionary.items():
            k = str(k)
            if not k.isidentifier():
                continue
            if v.__class__ is list:
                self.__setattr__(k, self._translate_list(v[:]))
            elif v.__class__ is dict:
                self.__setattr__(k, self.__class__(v))
            else:
                self.__setattr__(k, v)
        return self

    def __repr__(self):
        attrs = ', '.join(
            f'{attr}={value!r}'
            for attr, value in self.__dict__.items()
            if not attr.startswith('_')
        )
        return f'JSONObject({attrs})'

    __str__ = __repr__

    @classmethod
    def _translate_list(cls, items: list) -> list:
        for i in range(len(items)):
            item = items[i]
            if item.__class__ is list:
                items[i] = cls._translate_list(item)
            elif item.__class__ is dict:
                items[i] = cls(item)
        return items


class JsonObject(object):
    def __init__(self, dictionary: dict = None, **kvs):
        """
        将字典转化为对象（非递归）。

        :param dictionary: 包含数据的字典。请确保所有键都符合标识符命名要求，且均为字符串。
        :param kvs: 任意键值对。
        :return: JSONObject。
        :raise TypeError: 属性名称必须是字符串。
        """
        for k, v in {**(dictionary if dictionary else {}), **kvs}.items():
            k = str(k)
            if k.isidentifier():
                self.__setattr__(k, v)

    def __or__(self, dictionary: dict):
        for k, v in dictionary.items():
            k = str(k)
            if k.isidentifier():
                self.__setattr__(k, v)
        return self

    def __repr__(self):
        attrs = ', '.join(
            f'{attr}={value!r}'
            for attr, value in self.__dict__.items()
            if not attr.startswith('_')
        )
        return f'JsonObject({attrs})'

    __str__ = __repr__


def casting(
        mapper: Callable,
        raw,
        *errs: Exception,
        default=None,
):
    """
    使用mapper将raw转换为所需的值，当出现异常时返回default。

    默认捕获以下异常，可以通过errs参数追加更多异常：
        - ValueError
        - KeyboardInterrupt

    :param mapper: 类型转换器。如果转换器不可调用，将直接返回默认值。
    :param raw: 被转换的值。
    :param errs: 需要捕获的其它异常。应当提供可被 except 语句接受的值，否则会出现异常。
    :param default: 默认值。即使不提供也会默认返回 None 而不会抛出异常。
    :return: 转换后的值。如若捕获到特定异常将返回默认值。
    """
    if not callable(mapper):
        return default
    try:
        return mapper(raw)
    except (ValueError, KeyboardInterrupt) + errs:
        return default


class ReprMixin(object):
    class AttributeMeta:
        pass

    class TagMeta:
        pass

    def __repr__(self):
        kls = self.__class__.__name__
        pkv = self._obtain_pk()
        tags = self._obtain_tags()
        attrs = self._obtain_attrs()
        content = (
                (f'({pkv})' if pkv else '') +
                (f' {tags}' if tags else '') +
                (f' {attrs}' if attrs else '')
        )
        return f'<{kls}{content}>'

    @staticmethod
    def _fmt_value(value) -> str:
        if isinstance(value, str):
            return f'"{value}"'
        else:
            return str(value)

    def _obtain_attrs(self):
        def obtain():
            for attr, name in attributes.items():
                if attr.startswith('_'):
                    continue
                mapper = annotations.get(attr, None)
                value = getattr(self, attr)
                value = mapper(value) if callable(mapper) else value
                value = self._fmt_value(value)
                yield f'{name}={value}'

        attributes = self.AttributeMeta.__dict__
        annotations = attributes.get('__annotations__', {})
        return ', '.join(obtain())

    def _obtain_tags(self):
        def obtain():
            for attr, option in attributes.items():
                if attr.startswith('_'):
                    continue
                mapper = annotations.get(attr, None)
                value = getattr(self, attr)
                value = mapper(value) if callable(mapper) else value
                if isinstance(option, str) and value:
                    yield option
                elif isinstance(option, tuple):
                    yield option[bool(value)]
                elif isinstance(option, (list, dict)):
                    yield option[value]
                else:
                    pass

        attributes = self.TagMeta.__dict__
        annotations = attributes.get('__annotations__', {})
        return ' '.join(obtain())

    def _obtain_pk(self):
        if hasattr(self, 'pk'):
            return self._fmt_value(self.pk)
        if hasattr(self, 'id'):
            return self._fmt_value(self.id)
        return ''



from setuptools import setup

with open("./zeraora/__init__.py") as f:
    for line in f.readlines():
        if line.startswith("version"):
            delim = '"' if '"' in line else "'"
            version = line.split(delim)[1]
            break
    else:
        print("Can't find version! Stop Here!")
        exit(1)

with open('README.md', encoding='UTF-8') as f:
    long_description = f.read()

setup(
    name='Zeraora',
    version=version,
    url='https://github.com/aixcyi/zeraora',
    license='MIT',
    author='aixcyi',
    author_email='75880483+aixcyi@users.noreply.github.com',
    description="一个包含原创工具类及快捷函数的工具库。"
                "A original utility Python package.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Natural Language :: Chinese (Simplified)",
        "License :: OSI Approved :: MIT License",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
    project_urls={
        "Source": "https://github.com/aixcyi/zeraora",
        "Tracker": "https://github.com/aixcyi/zeraora/issues",
    },
    python_requires='>=3.7',
    packages=['zeraora'],
)



import os
from random import getrandbits
from time import time

from zeraora.charsets import BASE62, BASE64


def randbytes(n: int) -> bytes:
    """
    生成 n 个随机字节。

    此函数是对 Python 3.9 以前的 random.randbytes(n) 的替代。
    """
    assert n >= 0
    return getrandbits(n * 8).to_bytes(n, 'little')


def randb62(n: int) -> str:
    """
    生成 n 个 base62 随机字符。
    """
    return ''.join(BASE62[i % 62] for i in os.urandom(n))


def randb64(n: int) -> str:
    """
    生成 n 个 base64 随机字符。
    """
    return ''.join(BASE64[i >> 2] for i in os.urandom(n))


class SnowflakeWorker(object):
    STAMP = (1 << 41) - 1
    WORKER = (1 << 10) - 1
    SEQUENCE = (1 << 12) - 1
    EPOCH1970 = 0
    EPOCH2000 = 946656000
    EPOCH2020 = 1577808000

    def __init__(self, worker: int, *, epoch: int = EPOCH1970):
        """
        雪花ID生成的基本实现。

        雪花ID是一个64位整数，包含1位符号+41位时间戳+10位设备编号+12位毫秒内序号。

        :param worker: 设备编号。取值范围 0-1023。
        :param epoch: 时间戳起点（计算用的时间戳与标准时间戳的差，单位为毫秒）。
        """
        assert all([worker.__class__ is int,
                    epoch.__class__ is int])
        self._epoch = epoch
        self._stamp = 0
        self._worker = abs(worker) & self.WORKER
        self._sequence = 0

    def next_int(self) -> int:
        """
        获取下一个ID。

        :raise SnowflakeWorker.TimeRedirected: 时钟发生回拨。
        :raise SnowflakeWorker.SequenceOverflow: 同一毫秒内序号超出编码空间。
        """
        prev, curr = self._stamp, int(time() * 1000) - self._epoch

        if prev < curr:
            sequence = 1
        elif prev == curr:
            sequence = self._sequence + 1
        else:
            sequence = self.redirect()  # 时钟回拨

        if sequence & self.SEQUENCE != sequence:
            self.overflow()  # 序号上溢

        self._stamp = curr
        self._sequence = sequence
        return (self._stamp << 22) | (self._worker << 12) | sequence

    def redirect(self) -> int:
        raise self.TimeRedirected(self)

    def overflow(self):
        raise self.SequenceOverflow(self)

    def dump(self) -> dict:
        return {
            "stamp": self._stamp,
            "worker": self._worker,
            "sequence": self._sequence,
        }

    def load(self, data: dict, **kwargs):
        values = data | kwargs
        self._stamp = values.get('stamp', self._stamp)
        self._worker = values.get('worker', self._worker)
        self._sequence = values.get('sequence', self._sequence)

    class _Error(Exception):
        def __init__(self, parent: "SnowflakeWorker"):
            self.args = (
                parent._stamp,
                parent._worker,
                parent._sequence,
            )

    class TimeRedirected(_Error):
        def __str__(self):
            return 'Time redirected. ' \
                   '(stamp=%s, worker=%s, sequence=%s)' % self.args

    class SequenceOverflow(_Error):
        def __str__(self):
            return 'Snowflake ID sequence overflow. ' \
                   '(stamp=%s, worker=%s, sequence=%s)' % self.args


class SnowflakeMultiWorker(SnowflakeWorker):
    _objects_ = []

    def __new__(cls, worker: int, *, epoch: int = 0):
        """
        雪花ID生成的多例实现。

        雪花ID是一个64位整数，包含1位符号+41位时间戳+10位设备编号+12位毫秒内序号。

        :param worker: 设备编号。取值范围 0-1023。
        :param epoch: 时间戳起点（计算用的时间戳与标准时间戳的差，单位为毫秒）。
        :raise RuntimeError: 试图创建相同worker的类实例。
        """
        if worker in cls._objects_:
            raise RuntimeError(
                f'存在相同的 SnowflakeMultiWorker(worker={worker})'
            )
        cls._objects_.append(worker)
        return SnowflakeWorker.__new__(cls)


class SnowflakeSingleWorker(SnowflakeWorker):
    _instanced_ = False

    def __new__(cls, *, epoch: int = 0):
        """
        雪花ID生成的单例实现。

        雪花ID是一个64位整数，包含1位符号+41位时间戳+10位设备编号+12位毫秒内序号。
        其中设备编号从0开始，当发生时钟回拨时自动递增，因此总计容许 1023 次时钟回拨。

        :param epoch: 时间戳起点（计算用的时间戳与标准时间戳的差，单位为毫秒）。
        :raise RuntimeError: 试图创建第二个类实例。
        """
        if cls._instanced_ is True:
            raise RuntimeError('已实例化过 SnowflakeSingleWorker 。')
        cls._instanced_ = True
        return SnowflakeWorker.__new__(cls)

    def __init__(self, *, epoch: int = 0):
        super().__init__(0, epoch=epoch)

    def redirect(self) -> int:
        worker = self._worker + 1
        if worker & self.WORKER != worker:
            raise self.TimeRedirected(self)
        self._worker = worker
        return self._sequence



# print(''.join(map(chr, range(32, 127))))

# 按分类定义的字符集
DIGITS = '0123456789'
UPPERS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
LOWERS = 'abcdefghijklmnopqrstuvwxyz'
SYMBOL = r'''!"#$%&'()*+,-./:;<=>?@[\]^_`{|}~'''
SYMBOL_NORMAL = r"`-=[]\;',./"
SYMBOL_SHIFT = r'~!@#$%^&*()_+{}|:"<>?'
LETTERS = UPPERS + LOWERS
assert sorted(SYMBOL) == sorted(SYMBOL_NORMAL + SYMBOL_SHIFT)

# 去除易混淆字符后的字符集
DIGITS_SAFE = '23456789'
UPPERS_SAFE = 'ABCDEFGHJKLMNPQRSTUVWXYZ'
LOWERS_SAFE = 'abcdefghijklmnpqrstuvwxyz'
LETTERS_SAFE = UPPERS_SAFE + LOWERS_SAFE

# 按编码定义的字符集
BASE8 = '01234567'
BASE16 = '0123456789ABCDEF'
BASE36 = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
BASE62 = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
BASE64 = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz+/'
BASE64SAFE = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_'
OCTDIGITS = BASE8
HEXDIGITS = '0123456789ABCDEFabcdef'
